from qcp_omics.report_generation.report_step import report_step
from typing import TypeVar
from qcp_omics.utils.protocols import HasData
import plotly.express as px
import plotly.figure_factory as ff
import plotly.subplots as sp
import math


T = TypeVar("T", bound=HasData)


class VisualizationMixin:

    @staticmethod
    def _histograms(df):
        columns = df.columns

        if len(columns) == 0:
            return "<p>There are no columns in the forwarded dataset to generate histograms.</p>"

        rows = math.ceil(len(columns) / 5)

        fig = sp.make_subplots(rows=rows, cols=5, subplot_titles=columns)

        for i, col in enumerate(columns):
            dist = ff.create_distplot([df[col]], group_labels=[col])
            row_idx = i // 5 + 1
            col_idx = i % 5 + 1

            for trace in dist.data:
                fig.add_trace(trace, row=row_idx, col=col_idx)

        fig.update_layout(
            showlegend=False,
            height=300 * rows,
            width=1200
        )

        return fig.to_html(full_html=False)


    @staticmethod
    def _box_plots(df, columns):

        if len(columns) == 0:
            return "<p>There are no outliers</p>"

        rows = math.ceil(len(columns) / 5)

        fig = sp.make_subplots(rows=rows, cols=5, subplot_titles=columns)

        for i, col in enumerate(columns):
            box_fig = px.box(df, y=col)
            row_idx = i // 5 + 1
            col_idx = i % 5 + 1

            for trace in box_fig.data:
                fig.add_trace(trace, row=row_idx, col=col_idx)

        fig.update_layout(
            showlegend=False,
            height=300 * rows,
            width=1200
        )

        return fig.to_html(full_html=False)


    @staticmethod
    def _explained_variance(cum_var):
        fig = px.area(
            x=range(1, cum_var.shape[0] + 1),
            y=cum_var,
            labels={"x": "# Components", "y": "Explained Variance"}
        )
        return fig.to_html(full_html=False)


    @staticmethod
    def _pca_plot(df_pca, per_var):
        fig = px.scatter(df_pca,
                         x="PC1",
                         y="PC2",
                         height=800)
        return fig.to_html(full_html=False)


    @staticmethod
    def _heatmap(corr_df):
        fig = px.imshow(corr_df, width=800, height=800)
        return fig.to_html(full_html=False)



from qcp_omics.mixins.analysis_mixin import AnalysisMixin
from qcp_omics.mixins.visualization_mixin import VisualizationMixin
from qcp_omics.models.omics_data import OmicsData
from qcp_omics.mixins.preprocessing_mixin import PreprocessingMixin
from qcp_omics.mixins.qc_mixin import QCMixin


class ProteomicsData(OmicsData, QCMixin, PreprocessingMixin, AnalysisMixin, VisualizationMixin):
    pass


import functools
from typing import Tuple, Optional, Union, Literal

import pandas as pd


def report_step(snapshot: Optional[Literal["combined", "split", "numerical", "categorical"]] = None,
                output: bool = False
                ):
    def report_step_decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            result = func(self, *args, **kwargs)
            step_name = func.__name__

            data_snapshot: Optional[Union[Tuple[str, str], str]] = None
            # TODO: add error handling here
            if snapshot == "combined":
                data_snapshot = self._visualize_data_snapshot(self.data)
            elif snapshot == "split":
                data_snapshot = (self._visualize_data_snapshot(self.data_numerical),
                                 self._visualize_data_snapshot(self.data_categorical))
            elif snapshot == "numerical":
                data_snapshot = self._visualize_data_snapshot(self.data_numerical)
            elif snapshot == "categorical":
                data_snapshot = self._visualize_data_snapshot(self.data_categorical)

            if isinstance(result, pd.DataFrame):
                final_output = self._visualize_data_snapshot(result)
            elif isinstance(result, str):
                final_output = result
            elif isinstance(result, dict):
                final_output = {}
                for key, val in result.items():
                    if isinstance(val, pd.DataFrame):
                        final_output[key] = self._visualize_data_snapshot(val)
                    else:
                        final_output[key] = val
            else:
                final_output = result

            self.report_data.append({
                "step": step_name,
                "data_snapshot": data_snapshot,
                "data_snapshot_type": snapshot,
                "output": final_output if output else None
            })
            return result
        return wrapper
    return report_step_decorator



from qcp_omics.report_generation.report_step import report_step
from typing import TypeVar, Literal
from qcp_omics.utils.protocols import HasData


T = TypeVar("T", bound=HasData)


class AnalysisMixin:
    @report_step(output=True)
    def descriptive_statistics(self: T, method=None):
        if self.data_numerical is None or len(self.data_numerical.columns) == 0:
            return

        basic_stats = self.data_numerical.describe(include='all').T

        basic_stats['kurtosis'] = self.data_numerical.kurt()
        basic_stats['skewness'] = self.data_numerical.skew()

        return basic_stats


    @report_step(output=True)
    def pairwise_correlations_numerical(self: T, method: Literal["pearson", "spearman"] = "pearson"):
        if self.data_numerical is None or len(self.data_numerical) == 0:
            return

        corr_matrix = self.data_numerical.corr(method=method)
        heatmap = self._heatmap(corr_matrix)

        return {
            "corr_matrix": corr_matrix,
            "heatmap": heatmap
        }


    @report_step(output=True)
    def evaluate_distribution_features(self: T, method=None):
        if self.data_numerical is None or len(self.data_numerical) == 0:
            return

        hist_plots = self._histograms(self.data_numerical)
        return {
            "hist_plots": hist_plots
        }



import copy
import click
import typing as t

from .handle_execution import handle_execution
from qcp_omics.utils.utils import (
    prompt_already_run_steps,
    remove_previous_steps,
    prompt_steps_to_run,
    prompt_methods_if_needed,
    handle_json_input
)


ALL_STEPS: list[dict] = [
    {
      "step": "identify_missing_values"
    },
    {
      "step": "handle_missing_values",
    },
    {
      "step": "handle_outliers",
      "methods": ["IQR", "zscore"]
    },
    {
      "step": "split_train_test"
    },
    {
      "step": "split_numerical_categorical"
    },
    {
      "step": "scale_numerical_features",
      "methods": ["standard_scaler", "robust_scaler"]
    },
    {
      "step": "transform_numerical_features",
      "methods": ["box-cox", "log2"]
    },
    {
      "step": "descriptive_statistics"
    },
    {
      "step": "pairwise_correlations_numerical",
      "methods": ["pearson", "spearman"]
    },
    {
      "step": "evaluate_distribution_features"
    },
    {
      "step": "dimensionality_reduction"
    }
]
PREVIOUS_STEPS: list[str] = []

@click.command()
def interactive() -> None:

    click.echo("Welcome to QCP-Omics")

    # Basic prompts for dataset info
    dataset_type_options: list[str] = ["clinical", "genomics", "proteomics"]
    click.echo("\nWhat is the input dataset type:")
    for i, option in enumerate(dataset_type_options, 1):
        click.echo(f"{i}. {option}")
    choice = click.prompt("Choose one (1-3)", type=click.Choice(["1", "2", "3"]), show_choices=False)
    dataset_type = dataset_type_options[int(choice) - 1]

    dataset_path = click.prompt("\nPath to the source dataset", type=str)
    metadata_path = click.prompt("\nPath to the metadata file", type=str)
    output_path = click.prompt("\nPath to the directory where output should be saved", type=str)
    features_cols = click.confirm("\nAre features in columns and samples in rows in the input dataset?", default=True)
    en_header = click.confirm("\nAre all values in header and index in English?", default=True)
    is_raw = click.confirm("\nIs data raw (no processing applied yet)?", default=True)

    # Prepare the input dictionary for handle_execution
    cli_input: dict[str, t.Any] = {
        "dataset_type": dataset_type,
        "dataset_path": dataset_path,
        "metadata_path": metadata_path,
        "output_path": output_path,
        "report_path": output_path,
        "features_cols": features_cols,
        "en_header": en_header,
        "is_raw": is_raw
    }

    active_steps = copy.deepcopy(ALL_STEPS)

    if not is_raw:
        steps_already_run = prompt_already_run_steps(active_steps)
        PREVIOUS_STEPS.extend(steps_already_run)
        remove_previous_steps(active_steps, PREVIOUS_STEPS)

        click.echo("\nSteps already run (removed from the pipeline):")
        for step_name in steps_already_run:
            click.echo(f"  - {step_name}")

    steps_indices_to_run = prompt_steps_to_run(active_steps)
    steps_to_run: list[dict[str, t.Any]] = []

    for idx in steps_indices_to_run:
        step_dict = active_steps[idx]
        step_to_run = prompt_methods_if_needed(step_dict)
        steps_to_run.append(step_to_run)

    click.echo("\nSteps to be run (in this order):")
    for s in steps_to_run:
        if "method" in s:
            click.echo(f"  - {s['step']}  (method: {s['method']})")
        else:
            click.echo(f"  - {s['step']}")


    input_metadata = handle_json_input(cli_input["metadata_path"])

    cli_input["dtypes"] = input_metadata["dtypes"]
    cli_input["steps_to_run"] = steps_to_run

    handle_execution(cli_input)



from typing import Protocol, TypedDict, Optional
import pandas as pd

class HasData(Protocol):
    data: pd.DataFrame
    data_numerical: pd.DataFrame
    data_categorical: pd.DataFrame
    test_set: pd.DataFrame
    metadata: dict
    report_data: list[dict]



import click
from .handle_execution import handle_execution
from ..utils.utils import handle_json_input


@click.command()
@click.argument("input_path",
                type=click.Path(exists=True,
                                file_okay=True,
                                dir_okay=False,
                                readable=True))
def metadata(input_path) -> None:
    click.echo(f"Reading input from a metadata file: {input_path}")
    input_json = handle_json_input(input_path)
    handle_execution(input_json)



from abc import ABC
from typing import Optional
import pandas as pd

class OmicsData(ABC):
    def __init__(self, data: pd.DataFrame, metadata: dict) -> None:
        self.data = data
        self.data_numerical: Optional[pd.DataFrame] = None
        self.data_categorical: Optional[pd.DataFrame] = None
        self.test_set: Optional[pd.DataFrame] = None
        self.metadata = metadata
        self.report_data: list[dict] = []


    def __repr__(self):
        return f"<OmicsData(dataset_type: {self.metadata['dataset_type']})>"


    def transpose(self) -> None:
        if not self.metadata["features_cols"]:
            print("Transposing the dataset")
            self.data = self.data.T


    def map_dtypes(self) -> None:
        print("Mapping the dtypes from metadata with the dataset")
        dtype_mapping = self.metadata["dtypes"]
        for col, dtype in dtype_mapping.items():
            if col in self.data.columns:
                if dtype == "category":
                    self.data[col] = self.data[col].astype("category")
                elif dtype == "int":
                    self.data[col] = self.data[col].astype("int")
                elif dtype == "float":
                    self.data[col] = self.data[col].astype("float")


    @staticmethod
    def _visualize_data_snapshot(df: pd.DataFrame) -> str:
        html_table = df.to_html(classes="table table-striped table-bordered table-hover")
        return html_table


    def execute_steps(self) -> None:
        steps = self.metadata["steps_to_run"]
        for step in steps:
            step_impl = getattr(self, step["step"], None)
            method = step.get("method", None)
            if callable(step_impl):
                if method:
                    print(f"Executing step {step['step']} with {step['method']} method...")
                    step_impl(method=method)
                else:
                    print(f"Executing step {step['step']}...")
                    step_impl()
            else:
                print(f"Step {step['step']} is not recognised and will be skipped.")



from jinja2 import Environment, FileSystemLoader
import os


def generate_html_report(report_data, metadata, output_dir):
    script_dir = os.path.dirname(os.path.abspath(__file__))
    templates_dir = os.path.join(script_dir, "templates")

    if not os.path.isdir(output_dir):
        raise ValueError(f"The specified output path '{output_dir}' is not a valid directory.")

    env = Environment(loader=FileSystemLoader(templates_dir))
    template = env.get_template("report_template.jinja")

    html_content = template.render(data=report_data, metadata=metadata)

    output_file_path = os.path.join(output_dir, "report.html")

    with open(output_file_path, "w") as f:
        f.write(html_content)

    print(f"Report successfully generated at: {output_file_path}")



import numpy as np
import pandas as pd
from qcp_omics.report_generation.report_step import report_step
from typing import TypeVar, Any
from qcp_omics.utils.protocols import HasData
from sklearn.impute import SimpleImputer


T = TypeVar("T", bound=HasData)


class QCMixin:
    @staticmethod
    def _identify_missing_values(df: pd.DataFrame) -> dict[Any, float]:
        missing_values = df.isnull().mean() * 100
        filtered_missing = {col: pct for col, pct in missing_values.items() if pct > 0}
        sorted_missing = dict(sorted(filtered_missing.items(), key=lambda item: item[1], reverse=True))
        return sorted_missing


    def _impute_mean(self: T) -> None:
        imputer = SimpleImputer(strategy="mean")
        data_numerical = self.data.select_dtypes(include=["float", "int"])
        if len(data_numerical.columns) == 0:
            return
        imputed_values = imputer.fit_transform(data_numerical)
        imputed_df = pd.DataFrame(
            imputed_values,
            columns=data_numerical.columns,
            index=data_numerical.index
        )
        self.data[data_numerical.columns] = imputed_df
        for col in data_numerical.columns:
            self.data[col] = self.data[col].astype(data_numerical[col].dtype)


    def _impute_mode(self: T) -> None:
        imputer = SimpleImputer(strategy="most_frequent")
        data_categorical = self.data.select_dtypes(include=["category"])
        if len(data_categorical.columns) == 0:
            return
        imputed_values = imputer.fit_transform(data_categorical)
        imputed_df = pd.DataFrame(
            imputed_values,
            columns=data_categorical.columns,
            index=data_categorical.index,
        )
        self.data[data_categorical.columns] = imputed_df
        for col in data_categorical.columns:
            self.data[col] = self.data[col].astype('category')


    @staticmethod
    def _detect_outliers_iqr(df: pd.DataFrame) -> dict[str, list[tuple]]:
        outliers = {}
        for col in df.columns:
            q1 = df[col].quantile(0.25)
            q3 = df[col].quantile(0.75)
            iqr = q3 - q1
            lower_bound = q1 - 1.5 * iqr
            upper_bound = q3 + 1.5 * iqr
            outliers_mask = (df[col] < lower_bound) | (df[col] > upper_bound)
            col_outliers = df[col][outliers_mask]
            if not col_outliers.empty:
                outliers[col] = list(col_outliers.items())
        return outliers


    @staticmethod
    def _detect_outliers_zscore(df: pd.DataFrame, threshold: float = 3.0) -> dict[str, list[tuple]]:
        outliers = {}
        for col in df.columns:
            z_scores = np.abs((df[col] - df[col].mean()) / df[col].std())
            outliers_mask = z_scores > threshold
            col_outliers = df[col][outliers_mask]
            if not col_outliers.empty:
                outliers[col] = list(col_outliers.items())
        return outliers


    def _detect_outliers(self: T, data_numerical, method="iqr") -> dict[str, list[tuple]]:
        if method == "zscore":
            outliers = self._detect_outliers_zscore(data_numerical)
        else:
            outliers = self._detect_outliers_iqr(data_numerical)
        return outliers


    @report_step(output=True)
    def identify_missing_values(self: T, method=None) -> dict[Any, float]:
        return self._identify_missing_values(self.data)


    @report_step(snapshot="combined")
    def handle_missing_values(self: T, method=None):
        miss_cols = self._identify_missing_values(self.data)
        if not miss_cols:
            return
        for col, miss in miss_cols.items():
            if miss >= 30:
                print("here 2")
                self.data.drop(col, axis=1, inplace=True)
        self._impute_mode()
        self._impute_mean()


    @report_step(snapshot="combined", output=True)
    def handle_outliers(self: T, method="iqr") -> dict:
        data_numerical = self.data.select_dtypes(include=["float", "int"])

        outliers = self._detect_outliers(data_numerical, method=method)

        for col, outliers_list in outliers.items():
            median_value = self.data[col].median()
            for index, _ in outliers_list:
                self.data.at[index, col] = median_value

        boxplots = self._box_plots(data_numerical, list(outliers.keys()))

        return {
            "outliers": outliers,
            "boxplots": boxplots
        }
